"""
Placeholder module
"""


class Agent:
    """
    Agent placeholder stub for when smolagents isn't installed
    """

    def __init__(self, *args, **kwargs):
        """
        Raises an exception that smolagents isn't installed.
        """

        raise ImportError('smolagents is not available - install "agent" extra to enable')
